@@ -1,16 +1,28 @@
+import base64
+import json
+import time
+
+import jwt
 import requests
 
 
 class ModernMTException(Exception):
-    def __init__(self, status, type, message) -> None:
+    def __init__(self, status, type, message, metadata=None) -> None:
         super().__init__("(%s) %s" % (type, message))
+
         self.status = status
         self.type = type
         self.message = message
 
+        if metadata is not None:
+            self.metadata = metadata
+
 
 class ModernMT(object):
-    def __init__(self, api_key, platform="modernmt-python", platform_version="1.1.0", api_client=None) -> None:
+    def __init__(self, api_key, platform="modernmt-python", platform_version="1.4.0", api_client=None) -> None:
+        self.__batch_public_key = None
+        self.__batch_public_key_timestamp_sec = 0
+
         self.__base_url = "https://api.modernmt.com"
         self.__headers = {
             "MMT-ApiKey": api_key,
@@ -33,7 +45,7 @@ class ModernMT(object):
 
         res = self.__send("get", "/translate/detect", data=data)
 
-        if not isinstance(q, list):
+        if not isinstance(res, list):
             return DetectedLanguage(res)
 
         languages = []
@@ -66,10 +78,12 @@ class ModernMT(object):
                 data["format"] = options["format"]
             if "alt_translations" in options:
                 data["alt_translations"] = options["alt_translations"]
+            if "session" in options:
+                data["session"] = options["session"]
 
         res = self.__send("get", "/translate", data=data)
 
-        if not isinstance(q, list):
+        if not isinstance(res, list):
             return Translation(res)
 
         translations = []
@@ -78,6 +92,83 @@ class ModernMT(object):
 
         return translations
 
+    def batch_translate(self, webhook, source, target, q, hints=None, context_vector=None, options=None):
+        data = {
+            "webhook": webhook,
+            "target": target,
+            "q": q
+        }
+
+        if context_vector is not None:
+            data["context_vector"] = context_vector
+        if hints is not None:
+            if isinstance(hints, list):
+                hints = ",".join(map(str, hints))
+            data["hints"] = hints
+
+        if source is not None:
+            data["source"] = source
+
+        if options is not None:
+            if "project_id" in options:
+                data["project_id"] = options["project_id"]
+            if "multiline" in options:
+                data["multiline"] = options["multiline"]
+            if "format" in options:
+                data["format"] = options["format"]
+            if "alt_translations" in options:
+                data["alt_translations"] = options["alt_translations"]
+            if "metadata" in options:
+                data["metadata"] = options["metadata"]
+            if "session" in options:
+                data["session"] = options["session"]
+
+        headers = None
+        if options is not None and "idempotency_key" in options:
+            headers = {"x-idempotency-key": options["idempotency_key"]}
+
+        res = self.__send("post", "/translate/batch", data=data, headers=headers)
+
+        return res["enqueued"]
+
+    def handle_callback(self, body, signature):
+        if self.__batch_public_key is None:
+            self.__refresh_public_key()
+
+        if time.time() - self.__batch_public_key_timestamp_sec > 3600:
+            # noinspection PyBroadException
+            try:
+                self.__refresh_public_key()
+            except:
+                pass
+
+        jwt.decode(signature, self.__batch_public_key, algorithms=["RS256"])
+
+        if isinstance(body, str):
+            body = json.loads(body)
+
+        result = body["result"]
+        metadata = body.get("metadata", None)
+        error = result.get("error", None)
+
+        if error is not None:
+            raise ModernMTException(result["status"], error["type"], error["message"], metadata)
+
+        _data = result["data"]
+        if not isinstance(_data, list):
+            data = Translation(_data)
+        else:
+            data = []
+            for el in _data:
+                data.append(Translation(el))
+
+        return data, metadata
+
+    def __refresh_public_key(self):
+        data = self.__send("get", "/translate/batch/key")
+        self.__batch_public_key = base64.b64decode(data["publicKey"])
+        self.__batch_public_key_timestamp_sec = time.time()
+
     def get_context_vector(self, source, targets, text, hints=None, limit=None):
         data = {"source": source, "text": text, "targets": targets}
 
@@ -123,23 +214,42 @@ class ModernMT(object):
             else:
                 return None
 
-    def __send(self, method, endpoint, data=None, files=None, cls=None):
+    def me(self):
+        return self.__send("get", "/users/me", cls=User)
+
+    def qe(self, source, target, sentence, translation):
+        data = {"source": source, "target": target, "sentence": sentence, "translation": translation}
+        res = self.__send("get", "/qe", data=data)
+
+        if not isinstance(res, list):
+            return QualityEstimation(res)
+
+        qes = []
+        for el in res:
+            qes.append(QualityEstimation(el))
+
+        return qes
+
+    def __send(self, method, endpoint, data=None, files=None, cls=None, headers=None):
         url = self.__base_url + endpoint
 
-        headers = self.__headers
-        headers["X-HTTP-Method-Override"] = method
+        _headers = self.__headers
+        _headers["X-HTTP-Method-Override"] = method
+
+        if headers is not None:
+            _headers.update(headers)
 
         if files is None:
-            r = requests.post(url, headers=headers, json=data)
+            r = requests.post(url, headers=_headers, json=data)
         else:
-            r = requests.post(url, headers=headers, data=data, files=files)
+            r = requests.post(url, headers=_headers, data=data, files=files)
 
-        json = r.json()
+        _json = r.json()
         if r.status_code != requests.codes.ok:
             ex_type = "UnknownException"
             ex_msg = r.text
             try:
-                error = json["error"]
+                error = _json["error"]
                 ex_type, ex_msg = error["type"], error["message"]
             except KeyError:
                 pass
@@ -148,7 +258,7 @@ class ModernMT(object):
 
             raise ModernMTException(r.status_code, ex_type, ex_msg)
 
-        return json["data"] if cls is None else cls(json["data"])
+        return _json["data"] if cls is None else cls(_json["data"])
 
 
 class _MemoryServices(object):
@@ -189,15 +299,34 @@ class _MemoryServices(object):
     def delete(self, id):
         return self.__send("delete", "/memories/%s" % id, cls=Memory)
 
-    def add(self, id, source, target, sentence, translation, tuid=None):
-        data = {"source": source, "target": target, "sentence": sentence, "translation": translation}
+    def add(self, id, source, target, sentence, translation, tuid=None, session=None):
+        data = {
+            "source": source,
+            "target": target,
+            "sentence": sentence,
+            "translation": translation
+        }
+
         if tuid is not None:
             data["tuid"] = tuid
 
+        if session is not None:
+            data["session"] = session
+
         return self.__send("post", "/memories/%s/content" % id, data=data, cls=ImportJob)
 
-    def replace(self, id, tuid, source, target, sentence, translation):
-        data = {"tuid": tuid, "source": source, "target": target, "sentence": sentence, "translation": translation}
+    def replace(self, id, tuid, source, target, sentence, translation, session=None):
+        data = {
+            "tuid": tuid,
+            "source": source,
+            "target": target,
+            "sentence": sentence,
+            "translation": translation
+        }
+
+        if session is not None:
+            data["session"] = session
+
         return self.__send("put", "/memories/%s/content" % id, data=data, cls=ImportJob)
 
     def import_tmx(self, id, tmx, compression=None):
@@ -250,3 +379,39 @@ class ImportJob(_Model):
 class DetectedLanguage(_Model):
     def __init__(self, data) -> None:
         super().__init__(data, ["billedCharacters", "detectedLanguage"])
+
+
+class User(_Model):
+    def __init__(self, data) -> None:
+        super().__init__(data, [
+            "id",
+            "name",
+            "email",
+            "registrationDate",
+            "country",
+            "isBusiness",
+            "status"
+        ])
+        self.billingPeriod = BillingPeriod(data["billingPeriod"])
+
+
+class BillingPeriod(_Model):
+    def __init__(self, data) -> None:
+        super().__init__(data, [
+            "begin",
+            "end",
+            "chars",
+            "plan",
+            "planDescription",
+            "planForCatTool",
+            "amount",
+            "currency",
+            "currencySymbol"
+        ])
+
+
+class QualityEstimation(_Model):
+    def __init__(self, data) -> None:
+        super().__init__(data, [
+            "score"
+        ])
